/*
 * Copyright (C) 2012 Google, Inc.
 * Copyright (C) 2012 Square, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package dagger.internal;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

/**
 * A {@code Binding<T>} which contains contributors (other bindings marked with
 * {@code @Provides} {@code @OneOf}), to which it delegates provision
 * requests on an as-needed basis.
 */
public final class SetBinding<T> extends Binding<Set<T>> {

  public static <T> void add(Map<String, Binding<?>> bindings, String elementKey,
      Binding<?> binding) {
    @SuppressWarnings("unchecked")
    SetBinding<T> elementBinding = (SetBinding<T>) bindings.get(elementKey);
    if (elementBinding == null) {
      elementBinding = new SetBinding<T>(elementKey);
      bindings.put(elementBinding.provideKey, elementBinding);
    }
    elementBinding.contributors.add(Linker.scope(binding));
  }

  private final Set<Binding<?>> contributors = new LinkedHashSet<Binding<?>>();

  public SetBinding(String key) {
    super(key, null, false, null);
  }

  @Override public void attach(Linker linker) {
    for (Binding<?> contributor : contributors) {
      contributor.attach(linker);
    }
  }

  @SuppressWarnings("unchecked") // Bindings<T> are the only thing added to contributors.
  @Override public Set<T> get() {
    Set<T> result = new LinkedHashSet<T>(contributors.size());
    for (Binding<?> contributor : contributors) {
      result.add((T) contributor.get()); // Let runtime exceptions through.
    }
    return Collections.unmodifiableSet(result);
  }

  @Override public void getDependencies(
      Set<Binding<?>> getBindings, Set<Binding<?>> injectMembersBindings) {
    getBindings.addAll(contributors);
  }

  @Override public void injectMembers(Set<T> t) {
    throw new UnsupportedOperationException("Cannot inject into a multi-binder Set");
  }

  @Override public String toString() {
    return "SetBinding" + contributors;
  }
}
